@@ -37,6 +37,7 @@
 #include "match.h"
 #include "parse_config.h"
 #include "hal_interface.h" 
+#include "devices_list.h"
 
 static LibHalContext *hal_ctx;
 
@@ -51,14 +52,17 @@ void halevt_device_added(LibHalContext *ctx, const char *udi)
 /*
     DEBUG("New Device: %s", udi);
 */
+    DEBUG("New Device: %s", udi);
     while (current_insertion != NULL)
     {
-        if (halevt_true_tree(current_insertion->match, udi))
+        if (halevt_true_tree(current_insertion->match, udi, NULL))
         {
-            halevt_run_command(current_insertion->exec, udi);
+            halevt_run_command(current_insertion->exec, udi, NULL);
         }
         current_insertion = current_insertion->next;
     }
+    /* sync the devices list */
+    halevt_device_list_add_device (ctx, udi);
 }
 
 /** Invoked when a device is removed from the Global Device List. 
@@ -68,18 +72,41 @@ void halevt_device_added(LibHalContext *ctx, const char *udi)
 void halevt_device_removed(LibHalContext *ctx, const char *udi)
 {
     halevt_removal *current_removal = halevt_removal_root; 
+    halevt_device *removed_device;
    
 /*
     DEBUG("Device removed: %s", udi);
 */
+    DEBUG("Device removed: %s", udi);
 
-    while (current_removal != NULL)
+    /* use preferably the device from the devices list since the 
+     * properties will be right */
+    removed_device = halevt_device_list_find_device(udi);
+    if (removed_device != NULL)
     {
-        if (halevt_true_tree(current_removal->match, udi))
+        while (current_removal != NULL)
         {
-            halevt_run_command(current_removal->exec, udi);
+            if (halevt_true_tree(current_removal->match, NULL, removed_device))
+            {
+                halevt_run_command(current_removal->exec, udi, removed_device);
+            }
+            current_removal = current_removal->next;
+        }
+        /* sync the devices list */
+        halevt_device_list_remove_device(udi);
+    }
+    else
+    {
+        DEBUG("Cannot find device %s in halevt list", udi);
+        /* FIXME resync devices? */
+        while (current_removal != NULL)
+        {
+            if (halevt_true_tree(current_removal->match, udi, NULL))
+            {
+                halevt_run_command(current_removal->exec, udi, NULL);
+            }
+            current_removal = current_removal->next;
         }
-        current_removal = current_removal->next;
     }
 }
 
@@ -121,20 +148,29 @@ void halevt_device_property_modified(LibHalContext *ctx, const char *udi,
     while (current_property != NULL)
     {
        if (!strcmp(current_property->name, key) && 
-           halevt_true_tree(current_property->match, udi))
+           halevt_true_tree(current_property->match, udi, NULL))
        {
            current_action = current_property->action;
            while (current_action != NULL)
            {
-               if (halevt_property_matches(key, current_action->value, udi))
+               if (halevt_property_matches_udi(key, current_action->value, udi))
                {
-                   halevt_run_command(current_action->exec, udi);
+                   halevt_run_command(current_action->exec, udi, NULL);
                }
                current_action = current_action->next;
            }
        }
        current_property = current_property->next;
     }
+    /* now sync the properties in the devices list */
+    if (is_removed)
+    {
+        halevt_device_list_remove_property (udi, key);
+    }
+    else
+    {
+        halevt_device_list_set_property (udi, key);
+    }
 }
 
 
@@ -153,9 +189,9 @@ void halevt_device_condition(LibHalContext *ctx, const char *udi,
     while (current_condition != NULL)
     {
         if (!strcmp(current_condition->name, condition_name)
-            &&  halevt_true_tree(current_condition->match, udi))
+            &&  halevt_true_tree(current_condition->match, udi, NULL))
         {
-            halevt_run_command(current_condition->exec, udi);
+            halevt_run_command(current_condition->exec, udi, NULL);
         }
         current_condition = current_condition->next;
     }
@@ -233,6 +269,7 @@ void halevt_run_oninit()
    char **current_udi;
    int num_device;
    halevt_oninit *current_oninit;
+   halevt_device *device;
 
    dbus_error_init(&dbus_error);
    all_udi = libhal_get_all_devices (hal_ctx, &num_device, &dbus_error);
@@ -256,12 +293,14 @@ void halevt_run_oninit()
 
       while (current_oninit != NULL)
       {
-         if (halevt_true_tree(current_oninit->match, (*current_udi)))
+         if (halevt_true_tree(current_oninit->match, (*current_udi), NULL))
          {
-             halevt_run_command(current_oninit->exec, (*current_udi));
+             halevt_run_command(current_oninit->exec, (*current_udi), NULL);
          }
          current_oninit = current_oninit->next;
       }
+      /* construct the devices list */
+      device = halevt_device_list_add_device (hal_ctx, *current_udi);
       current_udi++;
    }
 }
@@ -270,7 +309,7 @@ void halevt_run_oninit()
  * return true if a property of a device specified by its udi
  * match a given value.
  */
-int halevt_property_matches (const char *property,
+int halevt_property_matches_udi (const char *property,
                          const char *value, const char *udi)
 {
    DBusError dbus_error;
@@ -278,7 +317,7 @@ int halevt_property_matches (const char *property,
 
    if ((udi == NULL) || (property == NULL))
    {
-      DEBUG("Warning: hal_property_matches called with a NULL value");
+      DEBUG("Warning: halevt_property_matches called with a NULL value");
       return 0;
    }
 
@@ -317,7 +356,7 @@ int halevt_property_matches (const char *property,
       {
           result = (libhal_device_get_property_int 
                (hal_ctx, udi, property, &dbus_error) == atoi(value));
-      }
+      } /* FIXME: LIBHAL_PROPERTY_TYPE_STRLIST */
       else if (type == LIBHAL_PROPERTY_TYPE_STRING)
       {
           char **cur_str;
@@ -364,24 +403,73 @@ int halevt_property_matches (const char *property,
 }
 
 /*
- * return true if a device matches a halevt_match
+ * return true if a property of a device specified by its device
+ * match a given value.
  */
-int halevt_matches (const halevt_match *match, const char *udi)
+int halevt_property_matches_device (const char *key,
+                         const char *value, const halevt_device *device)
+{
+   halevt_device_property *property;
+   char **values = NULL;
+
+   if ((device == NULL) || (key == NULL))
+   {
+      DEBUG("Warning: halevt_property_matches_device called with a NULL value");
+      return 0;
+   }
+   property = halevt_device_list_get_property (key, device);
+   if (property != NULL) { values = property->values ;}
+
+   if (values != NULL)
+   {
+       if (value == NULL) { return 1; }
+       while ((*values) != NULL)
+       {
+           if (!strcmp ((*values), value))
+           {
+               return 1;
+           }
+           values++;
+       }
+   }
+   return 0;
+}
+
+int halevt_match_device (const halevt_match *match, const halevt_device *device)
+{
+    const halevt_device *new_device = device;
+    const halevt_device_property *udi_property;
+    char **new_udi;
+    char **parent;
+
+    parent = match->parents;
+
+    while (*parent != NULL)
+    {
+       udi_property = halevt_device_list_get_property (*parent, new_device);
+       if (udi_property != NULL)
+       {
+          new_udi = udi_property->values;
+          new_device = halevt_device_list_find_device(new_udi[0]);
+          if (new_device == NULL) { return 0 ;}
+       }
+       else
+       {
+          return 0;
+       }
+       parent++;
+    }
+    return halevt_property_matches_device(match->name, match->value, new_device);
+}
+
+int halevt_match_udi (const halevt_match *match, const char *udi)
 {
     DBusError dbus_error;
     const char *new_udi = udi;
     char **parent;
 
-    if ((udi == NULL) || (match == NULL))
-    {
-        DEBUG("Warning: hal_matches was called with a NULL value");
-        return 0;
-    }
-
-
     dbus_error_init(&dbus_error);
 
-    if (!strcmp (match->name, "*")) { return 1; }
     parent = match->parents;
 
     while (*parent != NULL)
@@ -398,100 +486,279 @@ int halevt_matches (const halevt_match *match, const char *udi)
        halevt_check_dbus_error(&dbus_error);
        parent++;
     }
-    return halevt_property_matches(match->name, match->value, new_udi);
+    return halevt_property_matches_udi(match->name, match->value, new_udi);
 }
 
+/*
+ * return true if a device matches a halevt_match
+ */
+int halevt_matches (const halevt_match *match, const char *udi, 
+   const halevt_device *device)
+{
+    if (((udi == NULL) && (device == NULL)) || (match == NULL))
+    {
+        DEBUG("Warning: hal_matches was called with a NULL value");
+        return 0;
+    }
+
+    if (!strcmp (match->name, "*")) { return 1; }
+
+    if (udi != NULL) { return halevt_match_udi(match, udi) ;}
+    return (halevt_match_device (match, device));
+}
 
 /*
  * return the value of a property for a device specified by its udi
  */
-char *halevt_property_value (const char *property, const char *udi)
+char **halevt_udi_property_value (const char *property, const char *udi)
 {
-    char *value = NULL;
-    char tmp[256];
+    char **values = NULL;
 
     DBusError dbus_error;
 
     if ((udi == NULL) || (property == NULL))
     {
-        DEBUG("Warning: hal_property_value was called with a NULL value");
+        DEBUG("Warning: hal_udi_property_value was called with a NULL value");
         return NULL;
     }
-    if (!strcmp(property,"udi"))
-    {
-        return strdup(udi);
-    }
 
     dbus_error_init(&dbus_error);
 
-    if (libhal_device_property_exists
-        (hal_ctx, udi, property, &dbus_error))
+    if (libhal_device_property_exists(hal_ctx, udi, property, &dbus_error))
     {
         LibHalPropertyType type;
 
 	halevt_check_dbus_error(&dbus_error);
 
-	type = libhal_device_get_property_type
-            (hal_ctx, udi, property, &dbus_error);
+	type = libhal_device_get_property_type(hal_ctx, udi, property, &dbus_error);
 	halevt_check_dbus_error(&dbus_error);
 
-        if (type == DBUS_TYPE_STRING)
-        {
-            char *hal_value = libhal_device_get_property_string
-               (hal_ctx, udi, property, &dbus_error);
-            if (hal_value != NULL)
-            {
-                value = strdup(hal_value);
-                if (value != NULL) { libhal_free_string(hal_value); }
-            }
-        }
-        else if (type == DBUS_TYPE_BOOLEAN)
-        {
-            dbus_bool_t value_b = libhal_device_get_property_bool
-                (hal_ctx, udi, property, &dbus_error);
-            if (value_b == TRUE)
-            {
-                value = strdup("true");
-            }
-            else
-            {
-                value = strdup("false");
-            }
-        }
-        else if (type == DBUS_TYPE_INT32)
-        {
-            dbus_int32_t int_value = libhal_device_get_property_int
-                   (hal_ctx, udi, property, &dbus_error);
-            snprintf(tmp, 255, "%d", int_value);
-            tmp[255] = '\0';
-            value = strdup(tmp);
-        }
+        values = halevt_get_property_value(type, property, udi, &dbus_error);
+        
+    }
+    halevt_check_dbus_error(&dbus_error);
 
+    return values;
+}
 
-        else if (type == DBUS_TYPE_UINT64)
+char **halevt_property_value(const char *key, const char *udi,
+   const halevt_device* device)
+{
+    char **values;
+
+    if ((udi == NULL) || (key == NULL))
+    {
+        DEBUG("Warning: halevt_property_value was called with a NULL value");
+        return NULL;
+    }
+    if (!strcmp(key,"udi"))
+    {
+        char *new_udi = strdup(udi);
+        if (new_udi == NULL) { return NULL; }
+        values = malloc(2*sizeof(char*));
+        if (values == NULL)
         {
-            dbus_uint64_t uint_value = libhal_device_get_property_uint64
-                    (hal_ctx, udi, property,&dbus_error);
-            snprintf(tmp, 255, "%llu", uint_value);
-            tmp[255] = '\0';
-            value = strdup(tmp);
+            free (new_udi);
+            return NULL;
         }
-        else if (type == DBUS_TYPE_DOUBLE)
+        values[0] = new_udi;
+        values[1] = NULL;
+        return values;
+    }
+
+    if (device != NULL) 
+    { 
+        halevt_device_property *property = 
+             halevt_device_list_get_property(key, device);
+        return property->values;
+    }
+    return halevt_udi_property_value(key, udi);
+}
+
+char **halevt_duplicate_str_list(char** str_list)
+{
+   char **value = NULL;
+   char **cur_str;
+
+   if (str_list != NULL)
+   {
+       char *cur_val;
+       unsigned int total_value_nr;
+       char **cur_value;
+       total_value_nr = libhal_string_array_length(str_list);       
+
+       value = malloc ((total_value_nr+1) * sizeof(char*));
+       if (value == NULL)
+       { 
+           libhal_free_string_array (str_list);
+           return NULL;
+       }
+       *value = NULL;
+
+       cur_str = str_list;
+       cur_value = value;
+              
+       while ((*cur_str) != NULL)
+       {
+           cur_val = strdup(*cur_str);
+           if (cur_val == NULL)
+           { 
+               libhal_free_string_array (str_list);
+               cur_value = value;
+               while ((*cur_value) != NULL)
+               {
+                   free(*cur_value);
+                   cur_value++;
+               }
+               free(value);
+               return NULL;
+           }
+           *cur_value = cur_val;
+           cur_value++;
+           *cur_value = NULL;
+           cur_str++;
+       }
+       libhal_free_string_array (str_list);
+   }
+   return value;
+}
+
+/* cut and paste of halevt_get_property_value with libhal_psi_get_* instead of
+ * libhal_device_get_property_* */
+char **halevt_get_iterator_value(const LibHalPropertyType type, 
+  LibHalPropertySetIterator *iter)
+{
+    char **value = NULL;
+    char tmp[256];
+
+    if (type == LIBHAL_PROPERTY_TYPE_STRLIST)
+    {
+       char **str_list = libhal_psi_get_strlist (iter);
+       return halevt_duplicate_str_list(str_list);
+    }
+ 
+    value = malloc(2*sizeof(char *));
+    if (value == NULL) { return NULL; };
+    value[1] = NULL;
+
+    if (type == DBUS_TYPE_STRING)
+    {
+        char *hal_value = libhal_psi_get_string (iter);
+        if (hal_value != NULL)
         {
-            double dble_value = libhal_device_get_property_double
-                    (hal_ctx, udi, property,&dbus_error);
-            snprintf(tmp, 255, "%g", dble_value);
-            tmp[255] = '\0';
-            value = strdup(tmp);
+            value[0] = strdup(hal_value);
+            libhal_free_string(hal_value);
+        }
+    }
+    else if (type == DBUS_TYPE_BOOLEAN)
+    {
+        dbus_bool_t value_b = libhal_psi_get_bool (iter);
+        if (value_b == TRUE)
+        {
+            value[0] = strdup("true");
         }
         else
         {
-            DEBUG(_("Unhandled HAL type for property %s, device %s: %d!"), property, udi, type);
-            return NULL;
+            value[0] = strdup("false");
         }
-    }  
-    halevt_check_dbus_error(&dbus_error);
+    }
+    else if (type == DBUS_TYPE_INT32)
+    {
+        dbus_int32_t int_value = libhal_psi_get_int (iter);
+        snprintf(tmp, 255, "%d", int_value);
+        tmp[255] = '\0';
+        value[0] = strdup(tmp);
+    }
+    else if (type == DBUS_TYPE_UINT64)
+    {
+        dbus_uint64_t uint_value = libhal_psi_get_uint64 (iter);
+        snprintf(tmp, 255, "%llu", uint_value);
+        tmp[255] = '\0';
+        value[0] = strdup(tmp);
+    }
+    else if (type == DBUS_TYPE_DOUBLE)
+    {
+        double dble_value = libhal_psi_get_double (iter);
+        snprintf(tmp, 255, "%g", dble_value);
+        tmp[255] = '\0';
+        value[0] = strdup(tmp);
+    }
+    else
+    {
+        DEBUG(_("Unhandled HAL iterator type: %d!"), type);
+    }
+    return value;
+}
 
+char **halevt_get_property_value(LibHalPropertyType type, 
+  const char *property, const char *udi, DBusError *dbus_error_pointer)
+{
+    char **value = NULL;
+    char tmp[256];
+
+    if (type == LIBHAL_PROPERTY_TYPE_STRLIST)
+    {
+       char **str_list = libhal_device_get_property_strlist
+          (hal_ctx, udi, property, dbus_error_pointer);
+       
+       return halevt_duplicate_str_list(str_list);
+    }
+ 
+    value = malloc(2*sizeof(char *));
+    if (value == NULL) { return NULL; };
+    value[1] = NULL;
+
+    if (type == DBUS_TYPE_STRING)
+    {
+        char *hal_value = libhal_device_get_property_string
+            (hal_ctx, udi, property, dbus_error_pointer);
+        if (hal_value != NULL)
+        {
+            value[0] = strdup(hal_value);
+            libhal_free_string(hal_value);
+        }
+    }
+    else if (type == DBUS_TYPE_BOOLEAN)
+    {
+        dbus_bool_t value_b = libhal_device_get_property_bool
+            (hal_ctx, udi, property, dbus_error_pointer);
+        if (value_b == TRUE)
+        {
+            value[0] = strdup("true");
+        }
+        else
+        {
+            value[0] = strdup("false");
+        }
+    }
+    else if (type == DBUS_TYPE_INT32)
+    {
+        dbus_int32_t int_value = libhal_device_get_property_int
+               (hal_ctx, udi, property, dbus_error_pointer);
+        snprintf(tmp, 255, "%d", int_value);
+        tmp[255] = '\0';
+        value[0] = strdup(tmp);
+    }
+    else if (type == DBUS_TYPE_UINT64)
+    {
+        dbus_uint64_t uint_value = libhal_device_get_property_uint64
+                (hal_ctx, udi, property, dbus_error_pointer);
+        snprintf(tmp, 255, "%llu", uint_value);
+        tmp[255] = '\0';
+        value[0] = strdup(tmp);
+    }
+    else if (type == DBUS_TYPE_DOUBLE)
+    {
+        double dble_value = libhal_device_get_property_double
+                (hal_ctx, udi, property, dbus_error_pointer);
+        snprintf(tmp, 255, "%g", dble_value);
+        tmp[255] = '\0';
+        value[0] = strdup(tmp);
+    }
+    else
+    {
+        DEBUG(_("Unhandled HAL type for property %s, device %s: %d!"), property, udi, type);
+    }
     return value;
 }
 
@@ -499,7 +766,8 @@ char *halevt_property_value (const char *property, const char *udi)
  * run the command specified in halevt_exec, performing substitution
  * of hal properties.
  */
-int halevt_run_command(const halevt_exec *exec, char const *udi)
+int halevt_run_command(const halevt_exec *exec, char const *udi, 
+   const halevt_device *device)
 {
     char *argv[4];
     int string_size = 100;
@@ -508,6 +776,7 @@ int halevt_run_command(const halevt_exec *exec, char const *udi)
     GError *error = NULL;
     int string_index = 0;
     char *string;
+    char **values;
     int str_len;
     int hal_value_used;
 
@@ -522,14 +791,15 @@ int halevt_run_command(const halevt_exec *exec, char const *udi)
         hal_value_used = 0;
         if (exec->elements[string_index].hal_property)
         {
-            string = halevt_property_value (exec->elements[string_index].string, udi);
-            if (string == NULL)
+            values = halevt_property_value (exec->elements[string_index].string, udi, device);
+            if (values == NULL)
             {
-                DEBUG(_("Hal property %s in command '%s' not found"), exec->elements[string_index].string, exec->string);
+                DEBUG(_("Hal property %s in command '%s' not found (or OOM)"), exec->elements[string_index].string, exec->string);
                 string = "UNKNOWN";
             }
             else
             {
+                string = values[0];
                 hal_value_used = 1;
             }
         }
@@ -572,3 +842,5 @@ int halevt_run_command(const halevt_exec *exec, char const *udi)
     free(command);
     return 1;
 }
+
+
